"""Utils to package and publish.



The typical sequence of the methodic and paranoid could be something like this:

::

    python pack.py current-configs  # see what you got
    python pack.py increment-configs-version  # update (increment the version and write that in setup.cfg
    python pack.py current-configs-version  # see that it worked
    python pack.py current-configs  # ... if you really want to see the whole configs again (you're really paranoid)
    python pack.py run-setup  # see that it worked
    python pack.py twine-upload-dist  # publish
    # and then go check things work...



If you're crazy (or know what you're doing) just do

::

    python pack.py go

"""

import importlib
import subprocess
import pathlib
from setuptools import find_packages
import json
import re
import sys
import pkgutil
import os
from types import ModuleType
from pprint import pprint
from warnings import warn
from typing import (
    Union,
    Optional,
    Tuple,
    List,
    Literal,
)
from collections.abc import Mapping, Iterable, Generator, Sequence
from configparser import ConfigParser
from functools import partial

import argh

from wads.util import git
from wads.toml_util import (
    read_pyproject_toml,
    write_pyproject_toml,
    get_project_version,
    set_project_version,
    get_project_name,
)

CONFIG_FILE_NAME = "setup.cfg"
PYPROJECT_FILE_NAME = "pyproject.toml"
METADATA_SECTION = "metadata"
OPTIONS_SECTION = "options"
DFLT_OPTIONS = {
    "packages": "find:",
    "include_package_data": True,
    "zip_safe": False,
    "extras_require_testing": [],
}

pjoin = lambda *p: os.path.join(*p)
DOCSRC = "docsrc"
DFLT_PUBLISH_DOCS_TO = None  # 'github'


def _has_pyproject_toml(pkg_dir: str) -> bool:
    """Check if package uses pyproject.toml"""
    pkg_dir = os.path.realpath(pkg_dir)
    if pkg_dir.endswith(os.sep):
        pkg_dir = pkg_dir[:-1]
    pyproject_path = os.path.join(pkg_dir, PYPROJECT_FILE_NAME)
    return os.path.isfile(pyproject_path)


def _has_setup_cfg(pkg_dir: str) -> bool:
    """Check if package has setup.cfg"""
    pkg_dir = os.path.realpath(pkg_dir)
    if pkg_dir.endswith(os.sep):
        pkg_dir = pkg_dir[:-1]
    setup_cfg_path = os.path.join(pkg_dir, CONFIG_FILE_NAME)
    return os.path.isfile(setup_cfg_path)


def get_name_from_configs(pkg_dir, *, assert_exists=True):
    """Get name from local config file (pyproject.toml or setup.cfg)"""
    pkg_dir = _get_pkg_dir(pkg_dir)

    # Prefer pyproject.toml if it exists
    if _has_pyproject_toml(pkg_dir):
        name = get_project_name(pkg_dir)
    else:
        configs = read_configs(pkg_dir=pkg_dir)
        name = configs.get("name", None)

    if assert_exists:
        assert name is not None, "No name was found in configs"
    return name


def clog(condition, *args, log_func=pprint, **kwargs):
    if condition:
        log_func(*args, **kwargs)


def check_in(
    commit_message: str,
    *,
    work_tree: str = ".",
    git_dir: str | None = None,
    auto_choose_default_action: bool = False,
    bypass_docstring_validation: bool = False,
    bypass_tests: bool = False,
    bypass_code_formatting: bool = False,
    verbose: bool = False,
    pre_git_hooks: Sequence[str] = (),
):
    """Validate, normalize, stage, commit and push your local changes to a remote repository.

    :param commit_message: Your commit message
    :type commit_message: str
    :param work_tree: The relative or absolute path of the working directory. Defaults to '.'.
    :type work_tree: str, optional
    :param git_dir: The relative or absolute path of the git directory. If None, it will be taken to be "{work_tree}/.git/". Defaults to None.
    :type git_dir: str, optional
    :param auto_choose_default_action: Set to True if you don't want to be prompted and automatically select the default action. Defaults to False.
    :type auto_choose_default_action: bool, optional
    :param bypass_docstring_validation: Set to True if you don't want to check if a docstring exists for every module, class and function. Defaults to False.
    :type bypass_docstring_validation: bool, optional
    :param bypass_tests: Set to True if you don't want to run doctests and other tests. Defaults to False.
    :type bypass_tests: bool, optional
    :param bypass_code_formatting: Set to True if you don't want the code to be automatically formatted using axblack. Defaults to False.
    :type bypass_code_formatting: bool, optional
    :param verbose: Set to True if you want to log extra information during the process. Defaults to False.
    :type verbose: bool, optional
    :param pre_git_hooks: A sequence of git commands to run before the git commit. Defaults to ().
    """

    def ggit(command: str):
        r = git(command, work_tree=work_tree, git_dir=git_dir)
        clog(verbose, r, log_func=print)
        return r

    def print_step_title(step_title):
        if verbose:
            print()
            print(f"============= {step_title.upper()} =============")
        else:
            print(f"-- {step_title}")

    def abort():
        print("Aborting")
        sys.exit()

    def confirm(action, default):
        if auto_choose_default_action:
            return default
        return argh.interaction.confirm(action, default)

    def verify_current_changes():
        print_step_title("Verify current changes")
        if "nothing to commit, working tree clean" in ggit("status"):
            print("No changes to check in.")
            abort()

    def pull_remote_changes():
        print_step_title("Pull remote changes")
        ggit("stash")
        result = ggit("pull")
        ggit("stash apply")
        if result != "Already up to date." and not confirm(
            "Your local repository was not up to date, but it is now. Do you want to continue",
            default=True,
        ):
            abort()

    def validate_docstrings():
        def run_pylint(current_dir):
            import pylint.lint

            if os.path.exists(os.path.join(current_dir, "__init__.py")):
                result = pylint.lint.Run(
                    [
                        current_dir,
                        "--disable=all",
                        "--enable=C0114,C0115,C0116",
                    ],
                    do_exit=False,
                )
                if result.linter.stats["global_note"] < 10 and not confirm(
                    f"Docstrings are missing in directory {current_dir}. Do you want to continue",
                    default=True,
                ):
                    abort()
            else:
                for subdir in next(os.walk(current_dir))[1]:
                    run_pylint(os.path.join(current_dir, subdir))

        print_step_title("Validate docstrings")
        run_pylint(".")

    def run_tests():
        import pytest

        print_step_title("Run tests")
        args = ["--doctest-modules"]
        if verbose:
            args.append("-v")
        else:
            args.extend(["-q", "--no-summary"])
        result = pytest.main(args)
        if result == pytest.ExitCode.TESTS_FAILED and not confirm(
            "Tests have failed. Do you want to push anyway", default=False
        ):
            abort()
        elif result not in [
            pytest.ExitCode.OK,
            pytest.ExitCode.NO_TESTS_COLLECTED,
        ]:
            print("Something went wrong when running tests. Please check the output.")
            abort()

    def format_code():
        print_step_title("Format code")
        subprocess.run(
            "black --line-length=88 .",
            shell=True,
            check=True,
            capture_output=not verbose,
        )

    def commit_changes():
        print_step_title("Commit changes")
        ggit(f'commit --all --message="{commit_message}"')

    def push_changes():
        if not confirm(
            "Your changes have been commited. Do you want to push",
            default=True,
        ):
            abort()
        print_step_title("Push changes")
        ggit("push")

    try:
        for pre_git_hook in pre_git_hooks or []:
            ggit(pre_git_hook)
        verify_current_changes()
        pull_remote_changes()
        if not bypass_docstring_validation:
            validate_docstrings()
        if not bypass_tests:
            run_tests()
        if not bypass_code_formatting:
            format_code()
            verify_current_changes()  # check again after formatting
        commit_changes()
        push_changes()

    except subprocess.CalledProcessError:
        print("An error occured. Please check the output.")
        abort()

    print("Your changes have been checked in successfully!")


def goo(
    pkg_dir,
    commit_message,
    *,
    git_dir=None,
    auto_choose_default_action=False,
    bypass_docstring_validation=False,
    bypass_tests=False,
    bypass_code_formatting=False,
    verbose=False,
):
    """Validate, normalize, stage, commit and push your local changes to a remote repository.

    :param pkg_dir: The relative or absolute path of the working directory. Defaults to '.'.
    :type pkg_dir: str, optional
    :param commit_message: Your commit message
    :type commit_message: str
    :param git_dir: The relative or absolute path of the git directory. If None, it will be taken to be "{work_tree}/.git/". Defaults to None.
    :type git_dir: str, optional
    :param auto_choose_default_action: Set to True if you don't want to be prompted and automatically select the default action. Defaults to False.
    :type auto_choose_default_action: bool, optional
    :param bypass_docstring_validation: Set to True if you don't want to check if a docstring exists for every module, class and function. Defaults to False.
    :type bypass_docstring_validation: bool, optional
    :param bypass_tests: Set to True if you don't want to run doctests and other tests. Defaults to False.
    :type bypass_tests: bool, optional
    :param bypass_code_formatting: Set to True if you don't want the code to be automatically formatted using axblack. Defaults to False.
    :type bypass_code_formatting: bool, optional
    :param verbose: Set to True if you want to log extra information during the process. Defaults to False.
    :type verbose: bool, optional
    """

    return check_in(
        commit_message=commit_message,
        work_tree=pkg_dir,
        git_dir=git_dir,
        auto_choose_default_action=auto_choose_default_action,
        bypass_docstring_validation=bypass_docstring_validation,
        bypass_tests=bypass_tests,
        bypass_code_formatting=bypass_code_formatting,
        verbose=verbose,
    )


# TODO: Add a function that adds/commits/pushes the updated setup.cfg
# TODO: Include tests as first step
# TODO: blackify
# TODO: Git pull and make sure no conflicts before moving on...
def go(
    pkg_dir,
    *,
    version=None,
    publish_docs_to=DFLT_PUBLISH_DOCS_TO,
    verbose: bool = True,
    skip_git_commit: bool = False,
    answer_yes_to_all_prompts: bool = False,
    twine_upload_options_str: str = "",
    keep_dist_pkgs=False,
    commit_message="",
):
    """Update version, package and deploy:
    Runs in a sequence: increment_configs_version, update_setup_cfg, run_setup, twine_upload_dist

    :param version: The desired version (if not given, will increment the current version
    :param verbose: Whether to print stuff or not
    :param skip_git_commit: Whether to skip the git commit and push step
    :param answer_yes_to_all_prompts: If you do git commit and push, whether to ask confirmation after showing status

    """

    # TODO: Would like version to be decremented if the publication attempt fails!
    version = increment_configs_version(pkg_dir, version=version)
    update_setup_cfg(pkg_dir, verbose=verbose)
    run_setup(pkg_dir)
    twine_upload_dist(pkg_dir, options_str=twine_upload_options_str)

    if not keep_dist_pkgs:
        delete_pkg_directories(pkg_dir, verbose)

    if publish_docs_to:
        generate_and_publish_docs(pkg_dir, publish_docs_to)
    if not skip_git_commit:
        git_commit_and_push(
            pkg_dir,
            version=version,
            verbose=verbose,
            answer_yes_to_all_prompts=answer_yes_to_all_prompts,
            commit_message=commit_message,
        )


def git_commit_and_push(
    pkg_dir,
    *,
    version=None,
    verbose: bool = True,
    answer_yes_to_all_prompts: bool = False,
    commit_message="",
    what_to_add="*",
):
    def ggit(command):
        r = git(command, work_tree=pkg_dir)
        clog(verbose, r, log_func=print)

    ggit("status")

    if not answer_yes_to_all_prompts:
        answer = input("Should I do a 'git add *'? ([Y]/n): ")
        if answer and answer != "Y":
            print("Okay, I'll stop here.")
            return
    ggit(f"add {what_to_add}")

    ggit("status")  # show status again

    commit_command = f'commit -m "{version}: {commit_message}"'
    if not answer_yes_to_all_prompts:
        answer = input(f"Should I {commit_command} and push? ([Y]/n)")
        if answer and answer != "Y":
            print("Okay, I'll stop here.")
            return
    ggit(commit_command)
    ggit(f"push")


def generate_and_publish_docs(pkg_dir, *, publish_docs_to="github"):
    # TODO: Figure out epythet and wads relationship -- right now, there's a reflexive dependency
    from epythet import make_autodocs, make

    make_autodocs(pkg_dir)
    if publish_docs_to:
        make(pkg_dir, publish_docs_to)


PathStr = str
PkgName = str
PkgSpec = Union[PathStr, ModuleType, pathlib.Path, PkgName]


def delete_pkg_directories(pkg_dir: PathStr, verbose=True):
    from shutil import rmtree

    pkg_dir, pkg_dirname = extract_pkg_dir_and_name(pkg_dir)
    names_to_delete = ["dist", "build", f"{pkg_dirname}.egg-info"]
    for name_to_delete in names_to_delete:
        delete_dirpath = os.path.join(pkg_dir, name_to_delete)
        if os.path.isdir(delete_dirpath):
            if verbose:
                print(f"Deleting folder: {delete_dirpath}")
            rmtree(delete_dirpath)


def get_module_path(module: ModuleType) -> str:
    """
    Get the path to the directory containing the module's code file.

    >>> import os
    >>> get_module_path(os)  # doctest: +SKIP
    '/usr/lib/python3.8'
    >>> import sklearn  # doctest: +SKIP
    >>> get_module_path(sklearn)  # doctest: +SKIP
    '/usr/local/lib/python3.8/dist-packages/sklearn'
    >>> import local_package  # doctest: +SKIP
    >>> get_module_path(local_package)  # doctest: +SKIP
    '/home/user/projects/local_package/local_package'

    Read more: https://github.com/i2mint/wads/discussions/7#discussioncomment-9761632

    """
    if isinstance(module, ModuleType):
        loader = pkgutil.get_loader(module)
        if loader is None or loader.get_filename() is None:
            raise ValueError(f"Cannot find the filename for module {module.__name__}")
        return os.path.dirname(loader.get_filename())
    return module


def _get_pkg_dir_and_name(pkg_dir, validate=True):
    pkg_dir = os.path.realpath(pkg_dir)
    if pkg_dir.endswith(os.sep):
        pkg_dir = pkg_dir[:-1]
    pkg_name = os.path.basename(pkg_dir)
    if validate:
        if not os.path.isdir(pkg_dir):
            raise AssertionError(f"Directory {pkg_dir} wasn't found")
        if pkg_name not in os.listdir(pkg_dir):
            name_candidates = folders_that_have_init_py_files(pkg_dir)
            if not name_candidates:
                raise AssertionError(
                    f"pkg_dir={pkg_dir} doesn't itself contain a dir named {pkg_name}"
                )
            elif len(name_candidates) > 1:
                raise AssertionError(
                    f"pkg_dir={pkg_dir} contains multiple dirs with __init__.py files: "
                    f"{name_candidates}, so I don't know which to chose as the package name"
                )
            else:
                pkg_name = name_candidates[0]
    return pkg_dir, pkg_name


def extract_pkg_dir_and_name(
    pkg_spec: PkgSpec, *, validate: bool = True
) -> tuple[str, str]:
    """
    Extracts the pkg_dir and pkg_dirname from the input `pkg_spec`.
    Optionally validates the pkg_dir is actually one (has a pkg_name/__init__.py file)

    Also processes input to get a path from a pathlib.Path object, or a module object,
    or a module/package name string.

    `pkg_spec` can be an imported package (must be a locally developped package)
    whose name and containing directory is the same):

    >>> import wads
    >>> extract_pkg_dir_and_name(wads)  # doctest: +ELLIPSIS
    (.../wads', 'wads')

    You can also just specify the name of the package (it will be imported):

    >>> extract_pkg_dir_and_name('wads')  # doctest: +SKIP
    ('.../wads', 'wads')

    Or you can specify the path to the package directory explicitly:

    >>> extract_pkg_dir_and_name('/home/user/projects/wads')  # doctest: +SKIP
    ('/home/user/projects/wads', 'wads')

    """
    if isinstance(pkg_spec, pathlib.Path):
        pkg_spec = str(pkg_spec)
    # if pkg_spec has only alphanumeric characters, assume it's a package name and import it
    if isinstance(pkg_spec, str) and not os.path.isdir(pkg_spec) and pkg_spec.isalnum():
        pkg_spec = importlib.import_module(pkg_spec)
    # if pkg_spec is a module object, get the path to the module
    if isinstance(pkg_spec, ModuleType):
        module_dir = get_module_path(pkg_spec)
        # package dir is the parent of module dir
        # Note that this is only true for packages developped locally
        pkg_spec = os.path.dirname(module_dir)

    pkg_dir, pkg_name = _get_pkg_dir_and_name(pkg_spec)

    if validate:
        assert os.path.isdir(pkg_dir), f"Directory {pkg_dir} wasn't found"
        if not pkg_name in os.listdir(pkg_dir):
            raise AssertionError(
                f"pkg_dir={pkg_dir} doesn't itself contain a dir named {pkg_name}"
            )
        assert "__init__.py" in os.listdir(os.path.join(pkg_dir, pkg_name)), (
            f"pkg_dir={pkg_dir} contains a dir named {pkg_name}, "
            f"but that dir isn't a package (does not have a __init__.py"
        )

    return pkg_dir, pkg_name


def folders_that_have_init_py_files(pkg_dir: PathStr) -> list[str]:
    """
    Get a list of folders in the package directory that have an __init__.py file.

    >>> folders_that_have_init_py_files('/home/user/projects/wads')  # doctest: +SKIP
    ['wads', 'wads/util', 'wads/pack', 'wads/docs_gen']

    """
    return [
        d
        for d in os.listdir(pkg_dir)
        if os.path.isdir(os.path.join(pkg_dir, d))
        and "__init__.py" in os.listdir(os.path.join(pkg_dir, d))
    ]


def get_pkg_name(pkg_spec: PkgSpec, validate=True) -> PkgName:
    """
    Get the name of the package from a package name, module object or path.
    Optionally validates some naming rules.
    """
    pkg_dir, pkg_dirname = extract_pkg_dir_and_name(pkg_spec, validate=validate)
    configs_pkg_name = get_name_from_configs(pkg_dir)
    if validate:
        assert (
            pkg_dirname == configs_pkg_name
        ), f"({pkg_dirname=} and {configs_pkg_name=} were not the same"
    return configs_pkg_name


def _get_pkg_dir(pkg_spec: PkgSpec, validate=True) -> PathStr:
    """
    Get the path to the package directory from a package name, module object or path.
    """
    pkg_dir, _ = extract_pkg_dir_and_name(pkg_spec, validate=validate)
    return pkg_dir


def current_configs(pkg_dir):
    configs = read_configs(pkg_dir=_get_pkg_dir(pkg_dir))
    pprint(configs)


def current_configs_version(pkg_dir):
    pkg_dir = _get_pkg_dir(pkg_dir)

    # Prefer pyproject.toml if it exists
    if _has_pyproject_toml(pkg_dir):
        return get_project_version(pkg_dir)
    else:
        return read_configs(pkg_dir=pkg_dir).get("version")


# TODO: Both setup and twine are python. Change to use python objects directly.
def update_setup_cfg(pkg_dir, *, new_deploy=False, version=None, verbose=True):
    """Update setup.cfg or pyproject.toml.
    If version is not given, will ask pypi (via http request) what the current version
    is, and increment that.
    """
    pkg_dir = _get_pkg_dir(pkg_dir)
    configs = read_and_resolve_setup_configs(
        pkg_dir=_get_pkg_dir(pkg_dir),
        new_deploy=new_deploy,
        version=version,
    )
    pprint("\n{configs}\n")
    clog(verbose, pprint(configs))

    # Write to pyproject.toml if it exists, otherwise setup.cfg
    if _has_pyproject_toml(pkg_dir):
        # Update pyproject.toml with the resolved configs
        pyproject_data = read_pyproject_toml(pkg_dir)
        if "project" not in pyproject_data:
            pyproject_data["project"] = {}

        # Map key fields back to pyproject.toml format
        project = pyproject_data["project"]

        if "version" in configs:
            project["version"] = configs["version"]
        if "description" in configs:
            project["description"] = configs["description"]
        if "name" in configs:
            project["name"] = configs["name"]
        if "url" in configs:
            if "urls" not in project:
                project["urls"] = {}
            project["urls"]["Homepage"] = configs["url"]

        write_pyproject_toml(pkg_dir, pyproject_data)
    else:
        write_configs(pkg_dir=pkg_dir, configs=configs)


def set_version(pkg_dir, version):
    """Update version in config file (pyproject.toml and/or setup.cfg)"""
    pkg_dir = _get_pkg_dir(pkg_dir)
    assert isinstance(version, str), "version should be a string"

    # Update both files if both exist to keep them in sync
    has_pyproject = _has_pyproject_toml(pkg_dir)
    has_setup_cfg = _has_setup_cfg(pkg_dir)

    if has_pyproject:
        set_project_version(pkg_dir, version)

    if has_setup_cfg:
        configs = read_configs(pkg_dir)
        configs["version"] = version
        write_configs(pkg_dir=pkg_dir, configs=configs)


def increment_configs_version(
    pkg_dir,
    *,
    version=None,
):
    """Increment version in config file (pyproject.toml and/or setup.cfg)."""
    pkg_dir = _get_pkg_dir(pkg_dir)

    # Check which config files exist
    has_pyproject = _has_pyproject_toml(pkg_dir)
    has_setup_cfg = _has_setup_cfg(pkg_dir)

    # Determine the version to use
    if version is None:
        if has_pyproject:
            # Get current version from pyproject.toml
            current_version = get_project_version(pkg_dir)
            version = increment_version(current_version) if current_version else "0.0.1"
        else:
            # Get version from setup.cfg
            configs = read_configs(pkg_dir=pkg_dir)
            version = _get_version(
                pkg_dir, version=version, configs=configs, new_deploy=False
            )
            version = increment_version(version)

    # Update both files if both exist to keep them in sync
    if has_pyproject:
        set_project_version(pkg_dir, version)

    if has_setup_cfg:
        configs = read_configs(pkg_dir)
        configs["version"] = version
        write_configs(pkg_dir=pkg_dir, configs=configs)

    return version


def run_setup(pkg_dir):
    """Run ``python -m build`` (modern PEP 517 compliant build)"""
    print("--------------------------- build_output ---------------------------")
    pkg_dir = _get_pkg_dir(pkg_dir)
    original_dir = os.getcwd()
    os.chdir(pkg_dir)

    # Use python -m build (PEP 517 standard)
    # Falls back to legacy setup.py if build module not available
    try:
        build_output = subprocess.run(
            [sys.executable, "-m", "build"], capture_output=True, text=True
        )
        if build_output.returncode != 0:
            print(f"Build stderr: {build_output.stderr}")
            build_output.check_returncode()
        print(build_output.stdout)
    except (subprocess.CalledProcessError, FileNotFoundError) as e:
        # Fallback to legacy method if build module not installed
        print("Warning: 'build' module not found, falling back to setup.py")
        print("Consider installing: pip install build")
        build_output = subprocess.run(
            f"{sys.executable} setup.py sdist bdist_wheel".split(" ")
        )

    os.chdir(original_dir)


def twine_upload_dist(pkg_dir, *, options_str=None):
    """Publish to pypi. Runs ``python -m twine upload dist/*``"""
    print("--------------------------- upload_output ---------------------------")
    pkg_dir = _get_pkg_dir(pkg_dir)
    original_dir = os.getcwd()
    os.chdir(pkg_dir)
    # TODO: dist/*? How to publish just last one?
    if options_str:
        command = f"{sys.executable} -m twine upload {options_str} dist/*"
    else:
        command = f"{sys.executable} -m twine upload dist/*"
    # print(command)
    subprocess.run(command.split(" "))
    os.chdir(original_dir)
    # print(f"{upload_output.decode()}\n")


# -----------------------------------------------------------------------------
# setup.cfg reading and writing

# TODO: A lot of work done here to read setup.cfg. setup function apparently does it for you. How to use that?


# TODO: postprocess_ini_section_items and preprocess_ini_section_items: Add comma separated possibility?
# TODO: Find out if configparse has an option to do this processing alreadys
def postprocess_ini_section_items(items: Mapping | Iterable) -> Generator:
    r"""Transform newline-separated string values into actual list of strings (assuming that intent)

    >>> section_from_ini = {
    ...     'name': 'wads',
    ...     'keywords': '\n\tpackaging\n\tpublishing'
    ... }
    >>> section_for_python = dict(postprocess_ini_section_items(section_from_ini))
    >>> section_for_python
    {'name': 'wads', 'keywords': ['packaging', 'publishing']}

    """
    splitter_re = re.compile("[\n\r\t]+")
    if isinstance(items, Mapping):
        items = items.items()
    for k, v in items:
        if v.startswith("\n"):
            v = splitter_re.split(v[1:])
            v = [vv.strip() for vv in v if vv.strip()]
            v = [vv for vv in v if not vv.startswith("#")]  # remove commented lines
        yield k, v


# TODO: Find out if configparse has an option to do this processing already
def preprocess_ini_section_items(items: Mapping | Iterable) -> Generator:
    """Transform list values into newline-separated strings, in view of writing the value to a ini formatted section

    >>> section = {
    ...     'name': 'wads',
    ...     'keywords': ['documentation', 'packaging', 'publishing']
    ... }
    >>> for_ini = dict(preprocess_ini_section_items(section))
    >>> print('keywords =' + for_ini['keywords'])  # doctest: +NORMALIZE_WHITESPACE
    keywords =
        documentation
        packaging
        publishing

    """
    if isinstance(items, Mapping):
        items = items.items()
    for k, v in items:
        if isinstance(v, str) and not v.startswith('"') and "," in v:
            v = list(map(str.strip, v.split(",")))
        if isinstance(v, list):
            v = "\n\t" + "\n\t".join(v)

        yield k, v


def read_configs(
    pkg_dir: PathStr,
    postproc=postprocess_ini_section_items,
    section=METADATA_SECTION,
    *,
    verbose=False,
):
    assert isinstance(
        pkg_dir, PathStr
    ), "It doesn't look like pkg_dir is a path. Did you perhaps invert pkg_dir and postproc order"
    pkg_dir = _get_pkg_dir(pkg_dir)

    # Try pyproject.toml first (modern approach)
    pyproject_filepath = pjoin(pkg_dir, PYPROJECT_FILE_NAME)
    if os.path.isfile(pyproject_filepath):
        pyproject_data = read_pyproject_toml(pkg_dir)
        project_metadata = pyproject_data.get("project", {})

        # Convert pyproject.toml format to setup.cfg format for compatibility
        d = {}
        if project_metadata:
            # Map common fields
            if "name" in project_metadata:
                d["name"] = project_metadata["name"]
            if "version" in project_metadata:
                d["version"] = project_metadata["version"]
            if "description" in project_metadata:
                d["description"] = project_metadata["description"]
            if "urls" in project_metadata:
                urls = project_metadata["urls"]
                if isinstance(urls, dict):
                    # Try to get a root_url or homepage
                    if "Homepage" in urls:
                        d["url"] = urls["Homepage"]
                    elif "repository" in urls:
                        d["url"] = urls["repository"]
                    elif "Repository" in urls:
                        d["url"] = urls["Repository"]
            # Handle authors
            if "authors" in project_metadata:
                authors = project_metadata["authors"]
                if isinstance(authors, list) and authors:
                    # Join author names
                    d["author"] = ", ".join(
                        a.get("name", "") for a in authors if a.get("name")
                    )
            if "license" in project_metadata:
                license_info = project_metadata["license"]
                if isinstance(license_info, dict):
                    d["license"] = license_info.get("text", "")
                else:
                    d["license"] = str(license_info)
            if "keywords" in project_metadata:
                keywords = project_metadata["keywords"]
                if isinstance(keywords, list):
                    d["keywords"] = keywords
            if "dependencies" in project_metadata:
                d["install_requires"] = project_metadata["dependencies"]

        if verbose:
            print(f"Read configs from pyproject.toml: {d}")
        return d

    # Fall back to setup.cfg (legacy approach)
    config_filepath = pjoin(pkg_dir, CONFIG_FILE_NAME)
    c = ConfigParser()
    if os.path.isfile(config_filepath):
        c.read_file(open(config_filepath))
        if verbose:
            print(f"{section=}, {type(section)=}")
        try:
            d = c[section]
        except KeyError:
            d = {}
        if postproc:
            d = dict(postproc(d))
    else:
        d = {}
    return d


def write_configs(
    pkg_dir: PathStr,
    configs,
    preproc=preprocess_ini_section_items,
    dflt_options=DFLT_OPTIONS,
):
    assert isinstance(
        pkg_dir, PathStr
    ), "It doesn't look like pkg_dir is a path. Did you perhaps invert pkg_dir and configs order"
    pkg_dir = _get_pkg_dir(pkg_dir)
    config_filepath = pjoin(pkg_dir, CONFIG_FILE_NAME)
    c = ConfigParser()
    if os.path.isfile(config_filepath):
        c.read_file(open(config_filepath))

    metadata_dict = dict(preproc(configs))

    # Filter out None values and convert non-strings to strings for ConfigParser
    metadata_dict = {k: str(v) for k, v in metadata_dict.items() if v is not None}

    options = dict(dflt_options, **read_configs(pkg_dir, preproc, OPTIONS_SECTION))

    # TODO: Legacy. Reorg key to [section][key] mapping to avoid such ugly complexities.
    for k in [
        "install_requires",
        "install-requires",
        "packages",
        "zip_safe",
        "include_package_data",
    ]:
        if k in metadata_dict:
            if k not in options:
                options[k] = metadata_dict.pop(
                    k
                )  # get it out of metadata_dict and into options
            else:
                metadata_dict.pop(
                    k
                )  # if it's both in metadata and in options, just get it out of metadata

    # Handle nested sections like extras_require
    # Extract any keys that look like extras_require_* from both metadata and options
    extras_require = {}

    # Check metadata_dict first
    for key, value in list(metadata_dict.items()):
        if key.startswith("extras_require_"):
            extra_name = key.replace("extras_require_", "")
            extras_require[extra_name] = value
            metadata_dict.pop(key)

    # Check options dict too (since extras_require_testing might end up there via DFLT_OPTIONS)
    for key, value in list(options.items()):
        if key.startswith("extras_require_"):
            extra_name = key.replace("extras_require_", "")
            extras_require[extra_name] = value
            options.pop(key)

    c[METADATA_SECTION] = metadata_dict
    c[OPTIONS_SECTION] = options

    # Add extras_require section if we have any
    if extras_require:
        if "options.extras_require" not in c:
            c["options.extras_require"] = {}
        for extra_name, packages in extras_require.items():
            # Preprocess the packages (convert list to newline format if needed)
            if isinstance(packages, list):
                packages = "\n\t" + "\n\t".join(packages)
            c["options.extras_require"][extra_name] = packages

    with open(config_filepath, "w") as fp:
        c.write(fp)


# -----------------------------------------------------------------------------
# Versioning

from packaging.version import parse


def sorted_versions(strings: Iterable[str], version_patch_prefix=""):
    """
    Filter out and return version strings in (versioning) descending order.
    """
    version_pattern = re.compile(rf"^(\d+.){{2}}{version_patch_prefix}\d+$")

    # Filter and sort the versions in descending order
    sorted_versions = sorted(
        (
            x.replace(f"{version_patch_prefix}", "")
            for x in strings
            if version_pattern.match(x)
        ),
        key=parse,
        reverse=True,
    )
    return sorted_versions


def increment_version(version_str):
    version_nums = list(map(int, version_str.split(".")))
    version_nums[-1] += 1
    return ".".join(map(str, version_nums))


try:
    import requests

    requests_is_installed = True
except ModuleNotFoundError:
    requests_is_installed = False


def http_get_json(url, use_requests=requests_is_installed) -> dict | None:
    """Make ah http request to url and get json, and return as python dict"""

    if use_requests:
        import requests

        r = requests.get(url)
        if r.status_code == 200:
            return r.json()
        else:
            raise ValueError(f"response code was {r.status_code}")
    else:
        import urllib.request
        from urllib.error import HTTPError

        req = urllib.request.Request(url)
        try:
            r = urllib.request.urlopen(req)
            if r.code == 200:
                return json.loads(r.read())
            else:
                raise ValueError(f"response code was {r.code}")
        except HTTPError:
            return None  # to indicate (hopefully) that name doesn't exist
        except Exception:
            raise


PYPI_PACKAGE_JSON_URL = "https://pypi.python.org/pypi/{package}/json"


# TODO: Perhaps there's a safer way to analyze errors (and determine if the package exists or other HTTPError)
def versions_from_pypi(
    pkg_dir: PathStr,
    *,
    name: None | str = None,
    use_requests=requests_is_installed,
) -> str | None:
    """
    Return version of package on pypi.python.org using json.

    :param package: Name of the package
    :return: A version (string) or None if there was an exception (usually means there
    """
    name = get_pkg_name(pkg_dir)
    url = PYPI_PACKAGE_JSON_URL.format(package=name)

    try:
        pkg_info = http_get_json(url, use_requests=use_requests)
        releases = pkg_info.get("releases", [])
    except Exception as e:
        warn(f"Got an exception trying to get the versions from pypi: {e}")
        return []

    # keep only the versions that don't have yanked=True
    def yanked_release(release_info_list):
        return any(x.get("yanked", False) for x in release_info_list)

    releases = [k for k, v in releases.items() if not yanked_release(v)]
    return sorted_versions(releases)


def highest_pypi_version(
    pkg_dir: PathStr,
    *,
    name: None | str = None,
    use_requests=requests_is_installed,
) -> list[str]:
    """
    Return version of package on pypi.python.org using json.

    >>> highest_pypi_version('wads')  # doctest: +SKIP
    '0.1.19'

    :param package: Name of the package
    :return: A version (string) or None if there was an exception (usually means there
    """
    versions = versions_from_pypi(pkg_dir, name=name, use_requests=use_requests)
    if versions:
        return versions[0]
    # else: return None


def current_pypi_version(
    pkg_dir: PathStr,
) -> str | None:
    """
    Return version of package on pypi.python.org using json.

    >>> current_pypi_version('wads')  # doctest: +SKIP
    '0.1.19'

    """
    name = get_pkg_name(pkg_dir)
    url = PYPI_PACKAGE_JSON_URL.format(package=name)
    response = requests.get(url)
    if response.status_code == 200:
        data = response.json()
        return data["info"]["version"]
    else:
        # raise Exception(f"Failed to get information for package {name=}, {pkg_dir=}")
        return None


def next_version_for_package(
    pkg_dir: PathStr,
    name: None | str = None,
    version_if_current_version_none="0.0.1",
    use_requests=requests_is_installed,
) -> str:
    name = name or get_name_from_configs(pkg_dir=pkg_dir)
    current_version = current_pypi_version(name, use_requests=use_requests)
    if current_version is not None:
        return increment_version(current_version)
    else:
        return version_if_current_version_none


def _get_version(
    pkg_dir: PathStr,
    version,
    configs,
    name: None | str = None,
    new_deploy=False,
):
    version = version or configs.get("version", None)
    if version is None:
        try:
            if new_deploy:
                version = next_version_for_package(
                    pkg_dir, name
                )  # when you want to make a new package
            else:
                version = current_pypi_version(
                    pkg_dir, name
                )  # when you want to make a new package
        except Exception as e:
            print(
                f"Got an error trying to get the new version of {name} so will try to get the version from setup.cfg..."
            )
            print(f"{e}")
            version = configs.get("version", None)
            if version is None:
                raise ValueError(
                    f"Couldn't fetch the next version from PyPi (no API token?), "
                    f"nor did I find a version in setup.cfg (metadata section)."
                )
    return version


def versions_from_tags(pkg_spec: PkgSpec, version_patch_prefix: str = ""):
    pkg_dir, pkg_name = extract_pkg_dir_and_name(pkg_spec)
    tags = [x.strip() for x in git("tag", work_tree=pkg_dir).split("\n")]
    # Pattern to match versions with the patch prefix
    pattern = rf"^(\d+.){{2}}{version_patch_prefix}\d+$"
    # Filter and sort the versions in descending order
    sorted_versions = sorted(
        [
            x.replace(f"{version_patch_prefix}", "")
            for x in tags
            if re.match(pattern, x)
        ],
        key=parse,
        reverse=True,
    )
    return sorted_versions


def highest_tag_version(pkg_spec: PkgSpec, version_patch_prefix: str = ""):
    sorted_tag_versions = versions_from_tags(pkg_spec, version_patch_prefix)
    if len(sorted_tag_versions) > 0:
        return sorted_tag_versions[0]
    return None


def setup_cfg_version(pkg_spec: PkgSpec):
    """Get version from setup.cfg file."""
    pkg_dir, _ = extract_pkg_dir_and_name(pkg_spec)
    configs = read_configs(pkg_dir=pkg_dir)
    return configs.get("version", None)


def pyproject_toml_version(pkg_spec: PkgSpec):
    """Get version from pyproject.toml file."""
    pkg_dir, _ = extract_pkg_dir_and_name(pkg_spec)
    return get_project_version(pkg_dir)


def versions_from_different_sources(pkg_spec: PkgSpec):
    return {
        "tag": highest_tag_version(pkg_spec),
        "current_pypi": current_pypi_version(pkg_spec),
        "highest_not_yanked_pypi": highest_pypi_version(pkg_spec),
        "setup_cfg": setup_cfg_version(pkg_spec),
        "pyproject_toml": pyproject_toml_version(pkg_spec),
    }


ValidVersionSources = Literal[
    "tag", "current_pypi", "highest_not_yanked_pypi", "setup_cfg", "pyproject_toml"
]


def raise_error(msg, error_type=ValueError):
    raise error_type(msg)


def validate_versions(versions: dict, action_when_not_valid=raise_error) -> dict:
    """
    Validate versions from different sources.

    You get the versions input from the `versions_from_different_sources` function.

    :param versions: A dictionary with the versions from different sources
    :param action_when_not_valid: A function that will be called when the versions are not valid
        Default is to raise a ValueError with the error message.
        Another option is to print the error message, log it, or issue a warning.
    :return: The versions if they are valid
    """

    # TODO: Raise specific exceptions with what-to-do-about-it messages
    #   Tip: Write the instructions in a github wiki/discussion/issue and provide link

    error_msg = ""
    tag_version = versions.get("tag", None)
    if tag_version is not None and tag_version != versions["setup_cfg"]:
        error_msg += (
            f"Tag version ({tag_version}) is different "
            f"from setup.cfg's version: {versions['setup_cfg']}\n"
        )
    if versions["current_pypi"] != versions["highest_not_yanked_pypi"]:
        error_msg += (
            f"Current pypi version ({versions['current_pypi']}) is different "
            f"from the highest not yanked pypi version: {versions['highest_not_yanked_pypi']}\n"
        )
    if versions["current_pypi"] > versions["setup_cfg"]:
        error_msg += (
            f"Current pypi version ({versions['current_pypi']}) is higher "
            f"than setup.cfg's version: {versions['setup_cfg']}\n"
        )
    if error_msg:
        error_msg += (
            f"Please make sure the versions are consistent and then try again: \n"
            f"  {versions=}"
        )
        action_when_not_valid(error_msg)

    # but if all is well, return the versions:
    return versions


# -----------------------------------------------------------------------------


def read_and_resolve_setup_configs(
    pkg_dir: PathStr, *, new_deploy=False, version=None, assert_names=True
):
    """make setup params and call setup

    :param pkg_dir: Directory where the pkg is (which is also where the setup.cfg is)
    :param new_deploy: whether this setup for a new deployment (publishing to pypi) or not
    :param version: The version number to set this up as.
                    If not given will look at setup.cfg[metadata] for one,
                    and if not found there will use the current version (requesting pypi.org)
                    and bump it if the new_deploy flag is on
    """
    # read the config file (get a dict with it's contents)
    pkg_dir, pkg_dirname = _get_pkg_dir_and_name(pkg_dir)
    if assert_names:
        extract_pkg_dir_and_name(pkg_dir)

    configs = read_configs(pkg_dir)

    # parse out name and root_url
    # If URL is missing, try to infer it from git
    if "root_url" not in configs and "url" not in configs:
        try:
            from wads.populate import _get_pkg_url_from_pkg_dir

            git_url = _get_pkg_url_from_pkg_dir(pkg_dir)
            configs["url"] = git_url
            warn(
                f"No url or root_url found in configs. Inferred from git: {git_url}. "
                f"Consider adding it to your pyproject.toml or setup.cfg."
            )
        except Exception as e:
            raise ValueError(
                f"configs didn't have a root_url or url, and couldn't infer from git. "
                f"Please add a url field to your pyproject.toml [project.urls] section "
                f"or setup.cfg [metadata] section. Git error: {e}"
            )

    name = configs.get("name") or pkg_dirname
    if assert_names:
        assert (
            name == pkg_dirname
        ), f"config name ({name}) and pkg_dirname ({pkg_dirname}) are not equal!"

    if "root_url" in configs:
        root_url = configs["root_url"]
        if root_url.endswith(
            "/"
        ):  # yes, it's a url so it's always forward slash, not the systems' slash os.sep
            root_url = root_url[:-1]
        url = f"{root_url}/{name}"
    elif "url" in configs:
        url = configs["url"]
    else:
        raise ValueError(
            f"configs didn't have a root_url or url. It should have at least one of these!"
        )

    # Note: if version is not in config, version will be None,
    #  resulting in bumping the version or making it be 0.0.1 if the package is not found (i.e. first deploy)

    meta_data_dict = {k: v for k, v in configs.items()}

    # make the setup_kwargs
    setup_kwargs = dict(
        meta_data_dict,
        # You can add more key=val pairs here if they're missing in config file
    )

    version = _get_version(pkg_dir, version, configs, name, new_deploy)

    def text_of_readme_md_file():
        try:
            with open("README.md") as f:
                return f.read()
        except:
            return ""

    dflt_kwargs = dict(
        name=f"{name}",
        version=f"{version}",
        url=url,
        packages=find_packages(),
        include_package_data=True,
        platforms="any",
        # long_description=text_of_readme_md_file(),
        # long_description_content_type="text/markdown",
        description_file="README.md",
    )

    configs = dict(dflt_kwargs, **setup_kwargs)

    return configs


def _print_some_lines_of_code(file, file_contents, slice_):
    print(
        f"------------ lines {slice_.start} through {slice_.stop} of {file} ------------"
    )
    print("\n".join(file_contents.split("\n")[slice_]))
    print("---------------------------------------------------------")


def _slice_from_string(string):
    try:
        try:
            return slice(0, int(string))
        except ValueError:
            return slice(*map(int, string.split(":")))
    except ValueError:
        return None


def _ask_user_what_to_do_about_this_file(
    file, file_contents, dflt_n_lines_slice=20, start_here=0
):
    ask_again = lambda: _ask_user_what_to_do_about_this_file(
        file, file_contents, dflt_n_lines_slice, start_here
    )
    r = input(f"\nEnter a one liner doc for {file} then press enter:\n")
    r = r.strip()
    if r == "":
        start = start_here
        stop = start_here + dflt_n_lines_slice
        print_these_lines = slice(start, stop)
        start_here += (
            dflt_n_lines_slice  # so that next time r == '', we print new lines
        )
    else:
        print_these_lines = _slice_from_string(r)
    if print_these_lines is not None:
        _print_some_lines_of_code(file, file_contents, print_these_lines)
        # ... and then call the function again, recursively, until exit or some doc string is given
        return ask_again()
    else:
        if _equals_or_first_letter_of(r, "exit"):
            print("Exiting...")
            return "exit"
        if _equals_or_first_letter_of(r, "skip"):
            print("Skipping that one...")
            return "skip"
        if _equals_or_first_letter_of(r, "funcs"):
            _print_functions_and_classes(file, file_contents)
            return ask_again()
        if len(r) <= 5:
            print("---> Your docstring needs to be at least 5 characters")
            return ask_again()

        return r


def _print_functions_and_classes(file, file_contents):
    def gen():
        for line in file_contents.split("\n"):
            stripped_line = line.strip()
            if stripped_line.startswith("def ") or stripped_line.startswith("class "):
                yield line

    print(f"----------- funcs, classes, and methods of {file} -----------")
    print("\n".join(gen()))
    print("---------------------------------------------------------")


def _equals_or_first_letter_of(input_string, target_string):
    return input_string == target_string or (
        len(input_string) == 1 and input_string[0] == target_string[0]
    )


def process_missing_module_docstrings(
    *,
    pkg_dir: PathStr,
    action="input",
    exceptions=(),
    docstr_template='"""\n{user_input}\n"""\n',
):
    r"""
    Goes through modules of package, sees which ones don't have docstrings,
    and gives you the option to write one.

    The function will go through all .py files (except those mentioned in exceptions),
    check if there's a module docstring (that is, that the first non-white characters are
    triple-quotes (double or single)).

    What happens with that depends on the ``action`` argument,

    If ``action='list'``, those modules missing docstrings will be listed.
    If ``action='count'``, the count of those modules missing docstrings will be returned.

    If ``action='input'``, for every module you'll be given the option to enter a
    SINGLE LINE module docstring (though you could include multi-lines with \n).
    Just type the docstring you want and hit enter to go to the next module with missing docstring.
    Or, you can also:
        - type exit, or e: To exit this process
        - type skip, or s: To skip the module and go to the next
        - type funcs, or f: To see a print out of functions, classes, and methods
        - just hit enter, to get some lines of code (the first, then the next, etc.)
        - enter a number, or a number:number, to specify what lines you want to see printed
    Printing lines helps you write an informed module docstring

    :keyword module, docstr, docstrings, module doc strings
    """
    from dol import TextFiles, filt_iter

    pkg_dir, _ = _get_pkg_dir_and_name(pkg_dir)

    exceptions = set(exceptions)
    files = filt_iter(
        TextFiles(pkg_dir + "{}.py", max_levels=None),
        filt=exceptions.isdisjoint,
    )

    def files_and_contents_that_dont_have_docs():
        for file, file_contents in files.items():
            contents = file_contents.strip()
            if not contents.startswith('"""') and not contents.startswith("'''"):
                yield file, file_contents

    if action == "list":
        return [
            file for file, file_contents in files_and_contents_that_dont_have_docs()
        ]
    elif action == "count":
        return len(list(files_and_contents_that_dont_have_docs()))
    elif action == "input":
        were_no_files = True
        for file, file_contents in files_and_contents_that_dont_have_docs():
            were_no_files = False
            r = _ask_user_what_to_do_about_this_file(file, file_contents)
            if r == "exit":
                return
            elif r == "skip":
                continue

            files[file] = docstr_template.format(user_input=r) + file_contents

        if were_no_files:
            print("---> Seems like all your modules have docstrings! Congrads!")
            return True
    else:
        raise ValueError(f"Unknown action: {action}")


# -----------------------------------------------------------------------------


argh_kwargs = {
    "namespace": "pack",
    "functions": [
        generate_and_publish_docs,
        current_configs,
        increment_configs_version,
        current_configs_version,
        twine_upload_dist,
        read_and_resolve_setup_configs,
        update_setup_cfg,
        go,
        goo,
        check_in,
        get_name_from_configs,
        run_setup,
        current_pypi_version,
        extract_pkg_dir_and_name,
        git_commit_and_push,
        process_missing_module_docstrings,
    ],
    "namespace_kwargs": {
        "title": "Package Configurations",
        "description": "Utils to package and publish.",
    },
}


def main():
    import argh  # pip install argh

    argh.dispatch_commands(argh_kwargs.get("functions", None))


if __name__ == "__main__":
    main()



"""Utilities for reading and writing pyproject.toml files."""

import sys
import os
from typing import Any, Optional

# Use tomllib for Python 3.11+, tomli for earlier versions
if sys.version_info >= (3, 11):
    import tomllib
else:
    try:
        import tomli as tomllib
    except ImportError:
        tomllib = None

try:
    import tomli_w
except ImportError:
    tomli_w = None


def read_pyproject_toml(pkg_dir: str) -> dict[str, Any]:
    """
    Read pyproject.toml from the specified package directory.

    Args:
        pkg_dir: Path to the package directory

    Returns:
        Dictionary containing the parsed TOML data

    Raises:
        ImportError: If tomli/tomllib is not available
        FileNotFoundError: If pyproject.toml doesn't exist
    """
    if tomllib is None:
        raise ImportError(
            "tomli is required for Python < 3.11. Install with: pip install tomli"
        )

    pyproject_path = os.path.join(pkg_dir, "pyproject.toml")

    if not os.path.isfile(pyproject_path):
        return {}

    with open(pyproject_path, "rb") as f:
        return tomllib.load(f)


def write_pyproject_toml(pkg_dir: str, data: dict[str, Any]) -> None:
    """
    Write data to pyproject.toml in the specified package directory.

    Args:
        pkg_dir: Path to the package directory
        data: Dictionary to write as TOML

    Raises:
        ImportError: If tomli_w is not available
    """
    if tomli_w is None:
        raise ImportError(
            "tomli_w is required for writing TOML. Install with: pip install tomli_w"
        )

    pyproject_path = os.path.join(pkg_dir, "pyproject.toml")

    with open(pyproject_path, "wb") as f:
        tomli_w.dump(data, f)


def get_project_metadata(pkg_dir: str) -> dict[str, Any]:
    """
    Get the [project] section from pyproject.toml.

    Args:
        pkg_dir: Path to the package directory

    Returns:
        Dictionary containing project metadata
    """
    data = read_pyproject_toml(pkg_dir)
    return data.get("project", {})


def get_project_version(pkg_dir: str) -> Optional[str]:
    """
    Get the version from pyproject.toml.

    Args:
        pkg_dir: Path to the package directory

    Returns:
        Version string or None if not found
    """
    metadata = get_project_metadata(pkg_dir)
    return metadata.get("version")


def set_project_version(pkg_dir: str, version: str) -> None:
    """
    Set the version in pyproject.toml.

    Args:
        pkg_dir: Path to the package directory
        version: New version string
    """
    data = read_pyproject_toml(pkg_dir)

    if "project" not in data:
        data["project"] = {}

    data["project"]["version"] = version
    write_pyproject_toml(pkg_dir, data)


def get_project_name(pkg_dir: str) -> Optional[str]:
    """
    Get the project name from pyproject.toml.

    Args:
        pkg_dir: Path to the package directory

    Returns:
        Project name or None if not found
    """
    metadata = get_project_metadata(pkg_dir)
    return metadata.get("name")


def update_project_metadata(pkg_dir: str, **kwargs) -> None:
    """
    Update project metadata in pyproject.toml.

    Args:
        pkg_dir: Path to the package directory
        **kwargs: Metadata fields to update
    """
    data = read_pyproject_toml(pkg_dir)

    if "project" not in data:
        data["project"] = {}

    data["project"].update(kwargs)
    write_pyproject_toml(pkg_dir, data)


def update_project_url(pkg_dir: str, url: str, url_key: str = "Homepage") -> None:
    """
    Update or add a URL in the [project.urls] section of pyproject.toml.

    Args:
        pkg_dir: Path to the package directory
        url: The URL to set
        url_key: The key to use in the urls dict (default: "Homepage")
    """
    data = read_pyproject_toml(pkg_dir)

    if "project" not in data:
        data["project"] = {}

    if "urls" not in data["project"]:
        data["project"]["urls"] = {}

    data["project"]["urls"][url_key] = url
    write_pyproject_toml(pkg_dir, data)



"""
Comprehensive tests for populate and pack functionality with pyproject.toml.
"""

import subprocess
import sys
import tempfile
from pathlib import Path
import shutil

import pytest

from wads.populate import populate_pkg_dir, write_pyproject_configs
from wads.pack import (
    get_name_from_configs,
    current_configs_version,
    increment_configs_version,
    set_version,
    run_setup,
)
from wads.toml_util import read_pyproject_toml, get_project_version, get_project_name


def setup_git_repo(pkg_dir: Path, remote_url: str):
    """Helper function to initialize a git repo for testing"""
    subprocess.run(["git", "init"], cwd=pkg_dir, capture_output=True)
    subprocess.run(
        ["git", "config", "user.email", "test@example.com"],
        cwd=pkg_dir,
        capture_output=True,
    )
    subprocess.run(
        ["git", "config", "user.name", "Test User"],
        cwd=pkg_dir,
        capture_output=True,
    )
    subprocess.run(
        ["git", "remote", "add", "origin", remote_url],
        cwd=pkg_dir,
        capture_output=True,
    )


class TestPopulate:
    """Test the populate functionality"""

    def test_populate_creates_pyproject_toml(self):
        """Test that populate creates pyproject.toml"""
        with tempfile.TemporaryDirectory() as tmpdir:
            pkg_dir = Path(tmpdir) / "testpkg"
            pkg_dir.mkdir()
            setup_git_repo(pkg_dir, "https://github.com/test/testpkg")

            populate_pkg_dir(
                str(pkg_dir),
                description="Test package",
                root_url="https://github.com/test",
                author="Test Author",
                version="0.1.0",
                skip_ci_def_gen=True,  # Skip CI generation for simpler test
            )

            # Check that pyproject.toml was created
            pyproject_file = pkg_dir / "pyproject.toml"
            assert pyproject_file.exists(), "pyproject.toml should be created"

            # Check that the package directory was created
            assert (pkg_dir / "testpkg" / "__init__.py").exists()

            # Check that README was created
            assert (pkg_dir / "README.md").exists()

            # Check that LICENSE was created
            assert (pkg_dir / "LICENSE").exists()

    def test_populate_creates_correct_metadata(self):
        """Test that populate creates correct metadata in pyproject.toml"""
        with tempfile.TemporaryDirectory() as tmpdir:
            pkg_dir = Path(tmpdir) / "mypkg"
            pkg_dir.mkdir()
            setup_git_repo(pkg_dir, "https://github.com/myorg/mypkg")

            populate_pkg_dir(
                str(pkg_dir),
                description="My test package",
                root_url="https://github.com/myorg",
                author="John Doe",
                version="1.2.3",
                keywords=["test", "package"],
                install_requires=["requests", "click"],
                skip_ci_def_gen=True,
            )

            # Read the pyproject.toml
            data = read_pyproject_toml(str(pkg_dir))

            # Check metadata
            assert data["project"]["name"] == "mypkg"
            assert data["project"]["version"] == "1.2.3"
            assert data["project"]["description"] == "My test package"
            assert data["project"]["keywords"] == ["test", "package"]
            assert data["project"]["dependencies"] == ["requests", "click"]
            assert data["project"]["authors"] == [{"name": "John Doe"}]

    def test_write_pyproject_configs(self):
        """Test writing pyproject.toml from configs"""
        with tempfile.TemporaryDirectory() as tmpdir:
            pkg_dir = Path(tmpdir)

            configs = {
                "name": "testpkg",
                "version": "0.2.0",
                "description": "A test package",
                "url": "https://github.com/test/testpkg",
                "license": "apache-2.0",
                "keywords": ["test", "demo"],
                "author": "Test Author",
                "install_requires": ["numpy", "pandas"],
            }

            write_pyproject_configs(str(pkg_dir), configs)

            # Check that file was created and can be read
            data = read_pyproject_toml(str(pkg_dir))
            assert data["project"]["name"] == "testpkg"
            assert data["project"]["version"] == "0.2.0"


class TestPackWithPyprojectToml:
    """Test pack.py functions with pyproject.toml"""

    def test_get_name_from_configs(self):
        """Test getting package name from pyproject.toml"""
        with tempfile.TemporaryDirectory() as tmpdir:
            pkg_dir = Path(tmpdir) / "mypkg"
            pkg_dir.mkdir()
            setup_git_repo(pkg_dir, "https://github.com/test/mypkg")

            populate_pkg_dir(
                str(pkg_dir),
                description="Test package",
                root_url="https://github.com/test",
                skip_ci_def_gen=True,
            )

            name = get_name_from_configs(str(pkg_dir))
            assert name == "mypkg"

    def test_current_configs_version(self):
        """Test getting current version from pyproject.toml"""
        with tempfile.TemporaryDirectory() as tmpdir:
            pkg_dir = Path(tmpdir) / "versionpkg"
            pkg_dir.mkdir()
            setup_git_repo(pkg_dir, "https://github.com/test/versionpkg")

            populate_pkg_dir(
                str(pkg_dir),
                description="Test package",
                root_url="https://github.com/test",
                version="1.2.3",
                skip_ci_def_gen=True,
            )

            version = current_configs_version(str(pkg_dir))
            assert version == "1.2.3"

    def test_set_version(self):
        """Test setting version in pyproject.toml"""
        with tempfile.TemporaryDirectory() as tmpdir:
            pkg_dir = Path(tmpdir) / "verpkg"
            pkg_dir.mkdir()
            setup_git_repo(pkg_dir, "https://github.com/test/verpkg")

            populate_pkg_dir(
                str(pkg_dir),
                description="Test package",
                root_url="https://github.com/test",
                version="0.0.1",
                skip_ci_def_gen=True,
            )

            # Set new version
            set_version(str(pkg_dir), "2.0.0")

            # Verify it was updated
            version = get_project_version(str(pkg_dir))
            assert version == "2.0.0"

    def test_increment_configs_version(self):
        """Test incrementing version in pyproject.toml"""
        with tempfile.TemporaryDirectory() as tmpdir:
            pkg_dir = Path(tmpdir) / "incpkg"
            pkg_dir.mkdir()
            setup_git_repo(pkg_dir, "https://github.com/test/incpkg")

            populate_pkg_dir(
                str(pkg_dir),
                description="Test package",
                root_url="https://github.com/test",
                version="1.2.3",
                skip_ci_def_gen=True,
            )

            # Increment version
            new_version = increment_configs_version(str(pkg_dir))

            # Should be incremented to 1.2.4
            assert new_version == "1.2.4"

            # Verify it was updated in file
            version = get_project_version(str(pkg_dir))
            assert version == "1.2.4"


class TestBuild:
    """Test building packages"""

    def test_generated_package_can_be_built(self):
        """Test that a package generated by populate can be built"""
        with tempfile.TemporaryDirectory() as tmpdir:
            pkg_dir = Path(tmpdir) / "buildpkg"
            pkg_dir.mkdir()
            setup_git_repo(pkg_dir, "https://github.com/test/buildpkg")

            populate_pkg_dir(
                str(pkg_dir),
                description="A buildable package",
                root_url="https://github.com/test",
                version="0.1.0",
                install_requires=["requests"],
                skip_ci_def_gen=True,
            )

            # Try to build the package
            try:
                run_setup(str(pkg_dir))

                # Check that dist directory was created
                dist_dir = pkg_dir / "dist"
                assert dist_dir.exists(), "dist directory should be created"

                # Check that wheel and sdist were created
                dist_files = list(dist_dir.glob("*"))
                assert len(dist_files) > 0, "Build should create distribution files"

                # Should have at least a .tar.gz and .whl
                extensions = {f.suffix for f in dist_files}
                assert ".whl" in extensions or ".gz" in extensions

            except subprocess.CalledProcessError as e:
                # Print output for debugging
                print(f"Build failed: {e}")
                pytest.fail(f"Package build failed: {e}")

    def test_built_package_passes_twine_check(self):
        """Test that built packages pass twine check"""
        with tempfile.TemporaryDirectory() as tmpdir:
            pkg_dir = Path(tmpdir) / "twinepkg"
            pkg_dir.mkdir()
            setup_git_repo(pkg_dir, "https://github.com/test/twinepkg")

            populate_pkg_dir(
                str(pkg_dir),
                description="A package for twine checking",
                root_url="https://github.com/test",
                version="0.1.0",
                skip_ci_def_gen=True,
            )

            # Build the package
            run_setup(str(pkg_dir))

            # Run twine check
            result = subprocess.run(
                [sys.executable, "-m", "twine", "check", "dist/*"],
                cwd=pkg_dir,
                capture_output=True,
                text=True,
            )

            # Skip if twine not installed
            if result.returncode != 0 and "No module named twine" in result.stderr:
                pytest.skip("twine not installed")

            # Twine check should pass
            assert (
                result.returncode == 0
            ), f"Twine check failed: {result.stdout}\n{result.stderr}"
            assert "PASSED" in result.stdout or result.returncode == 0


class TestCompatibility:
    """Test backward compatibility with setup.cfg"""

    def test_can_read_legacy_setup_cfg(self):
        """Test that pack can still read legacy setup.cfg files"""
        # This test would require a legacy setup.cfg file
        # For now, just ensure the functions don't crash when pyproject.toml is missing
        pass


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



"""
Quick Reference Guide for wads.migration
=========================================

BASIC SETUP.CFG MIGRATION
--------------------------

From file:
    >>> from wads.migration import migrate_setuptools_to_hatching  # doctest: +SKIP
    >>> pyproject = migrate_setuptools_to_hatching('setup.cfg')  # doctest: +SKIP
    >>> with open('pyproject.toml', 'w') as f:  # doctest: +SKIP
    ...     f.write(pyproject)

From string:
    >>> setup_cfg_text = "[metadata]\\nname=myproj\\nversion=1.0.0"  # doctest: +SKIP
    >>> pyproject = migrate_setuptools_to_hatching(setup_cfg_text)  # doctest: +SKIP

From dict:
    >>> cfg = {'metadata': {'name': 'myproj', 'version': '1.0.0'}}  # doctest: +SKIP
    >>> pyproject = migrate_setuptools_to_hatching(cfg)  # doctest: +SKIP


WITH DEFAULTS FOR MISSING FIELDS
---------------------------------

    >>> pyproject = migrate_setuptools_to_hatching(  # doctest: +SKIP
    ...     'setup.cfg',
    ...     defaults={
    ...         'description': 'My awesome project',
    ...         'url': 'https://github.com/user/project',
    ...         'license': 'MIT'
    ...     }
    ... )


CUSTOM TRANSFORMATION RULES
----------------------------

    >>> custom_rules = {  # doctest: +SKIP
    ...     'project.name': lambda cfg: cfg['metadata']['name'].upper(),
    ...     'project.version': lambda cfg: cfg['metadata']['version'],
    ...     'project.description': lambda cfg: "Custom description",
    ... }
    >>> pyproject = migrate_setuptools_to_hatching('setup.cfg', rules=custom_rules)  # doctest: +SKIP


ERROR HANDLING
--------------

    >>> from wads.migration import MigrationError  # doctest: +SKIP
    >>> try:  # doctest: +SKIP
    ...     pyproject = migrate_setuptools_to_hatching(incomplete_cfg)
    ... except MigrationError as e:
    ...     print(f"Missing: {e}")
    ...     # Provide defaults and retry


CI MIGRATION
------------

From file:
    >>> from wads.migration import migrate_github_ci_old_to_new
    >>> new_ci = migrate_github_ci_old_to_new('.github/workflows/ci.yml')
    >>> with open('.github/workflows/ci_new.yml', 'w') as f:  # doctest: +SKIP
    ...     f.write(new_ci)

With defaults:
    >>> new_ci = migrate_github_ci_old_to_new(  # doctest: +SKIP
    ...     'ci.yml',
    ...     defaults={'project_name': 'myproject'}
    ... )


COMPLETE MIGRATION WORKFLOW
----------------------------

Migrate setup.cfg and CI files:

    >>> from pathlib import Path  # doctest: +SKIP
    >>> from wads.migration import (  # doctest: +SKIP
    ...     migrate_setuptools_to_hatching,
    ...     migrate_github_ci_old_to_new
    ... )
    >>> project = Path('/path/to/project')  # doctest: +SKIP
    >>> pyproject = migrate_setuptools_to_hatching(str(project / 'setup.cfg'))  # doctest: +SKIP
    >>> (project / 'pyproject.toml').write_text(pyproject)  # doctest: +SKIP
    >>> old_ci = project / '.github/workflows/ci.yml'  # doctest: +SKIP
    >>> if old_ci.exists():  # doctest: +SKIP
    ...     new_ci = migrate_github_ci_old_to_new(str(old_ci))
    ...     (project / '.github/workflows/ci_new.yml').write_text(new_ci)


AVAILABLE TRANSFORMATION RULES
-------------------------------

The following fields are automatically extracted from setup.cfg:

    project.name                     - Project name
    project.version                  - Version number
    project.description              - Description
    project.url                      - Homepage URL
    project.license                  - License identifier
    project.keywords                 - Keywords list
    project.authors                  - Author information
    project.dependencies             - install_requires → dependencies
    project.optional-dependencies    - extras_require
    project.scripts                  - console_scripts entry points


ADDING NEW RULES
----------------

    >>> from wads.migration import setup_cfg_to_pyproject_toml_rules
    >>>
    >>> def _rule_custom_field(cfg: dict) -> str:
    ...     return cfg.get('metadata', {}).get('custom', 'default')
    >>>
    >>> setup_cfg_to_pyproject_toml_rules['project.custom'] = _rule_custom_field


COMMON PATTERNS
---------------

Check if migration needed:
    >>> from pathlib import Path  # doctest: +SKIP
    >>> project = Path('/path/to/project')  # doctest: +SKIP
    >>> needs_migration = (  # doctest: +SKIP
    ...     (project / 'setup.cfg').exists() and
    ...     not (project / 'pyproject.toml').exists()
    ... )

Batch migration:
    >>> projects = [Path(p) for p in ['proj1', 'proj2', 'proj3']]  # doctest: +SKIP
    >>> for proj in projects:  # doctest: +SKIP
    ...     if (proj / 'setup.cfg').exists():
    ...         pyproject = migrate_setuptools_to_hatching(str(proj / 'setup.cfg'))
    ...         (proj / 'pyproject.toml').write_text(pyproject)

Dry-run (show what would be done):
    >>> from wads.migration import migrate_setuptools_to_hatching
    >>> pyproject = migrate_setuptools_to_hatching('setup.cfg')
    >>> print(pyproject[:500])  # doctest: +ELLIPSIS
    [build-system]
    requires = [
        "hatchling",
    ]
    ...


For complete documentation, see MIGRATION.md
"""



"""Licensing"""

from wads import licenses_json_path
import json
import os
import re
from datetime import datetime
from typing import Optional
from urllib.parse import urlparse


def resolve_author(
    author: Optional[str] = None,
    pyproject_authors: Optional[list] = None,
    url: Optional[str] = None,
) -> str:
    """
    Resolve the author name following a priority chain.

    Priority order:
    1. If author is explicitly provided, use it
    2. If pyproject_authors is provided, use the first author's name
    3. If url is a GitHub URL, extract the org/user from it
    4. If WADS_DFLT_AUTHOR env variable is set, use it
    5. Return placeholder if nothing else works

    Args:
        author: Explicitly provided author name
        pyproject_authors: List of author dicts from pyproject.toml
        url: Project URL (e.g., GitHub URL)

    Returns:
        Resolved author name or placeholder

    >>> resolve_author(author='John Doe')
    'John Doe'
    >>> resolve_author(pyproject_authors=[{'name': 'Jane Smith'}])
    'Jane Smith'
    >>> resolve_author(url='https://github.com/myorg/myrepo')
    'myorg'
    """
    # 1. Explicit author
    if author:
        return author

    # 2. Authors from pyproject.toml
    if pyproject_authors and isinstance(pyproject_authors, list):
        if pyproject_authors:
            first_author = pyproject_authors[0]
            if isinstance(first_author, dict) and "name" in first_author:
                return first_author["name"]
            elif isinstance(first_author, str):
                return first_author

    # 3. Extract from GitHub URL
    if url:
        # Handle both https and git URLs
        github_patterns = [
            r"https://github\.com/([^/]+)",
            r"git@github\.com:([^/]+)",
            r"http://github\.com/([^/]+)",
        ]
        for pattern in github_patterns:
            match = re.search(pattern, url)
            if match:
                return match.group(1)

    # 4. Environment variable
    env_author = os.environ.get("WADS_DFLT_AUTHOR")
    if env_author:
        return env_author

    # 5. Return placeholder
    return "[fullname]"


def substitute_license_placeholders(
    license_text: str,
    author: Optional[str] = None,
    year: Optional[int] = None,
) -> str:
    """
    Replace placeholders in license text with actual values.

    Common placeholders:
    - [yyyy], [year] -> current year
    - [fullname], [name of copyright owner] -> author name
    - (c) -> © (optional enhancement)

    Args:
        license_text: License text with placeholders
        author: Author/copyright owner name (if None, placeholder remains)
        year: Year for copyright (if None, uses current year)

    Returns:
        License text with substituted values

    >>> text = "Copyright [yyyy] [fullname]"
    >>> result = substitute_license_placeholders(text, author='John', year=2025)
    >>> result
    'Copyright 2025 John'
    """
    if not license_text:
        return license_text

    # Use current year if not provided
    if year is None:
        year = datetime.now().year

    result = license_text

    # Replace year placeholders
    year_patterns = [
        (r"\[yyyy\]", str(year)),
        (r"\[year\]", str(year)),
        (r"\(yyyy\)", str(year)),
    ]

    for pattern, replacement in year_patterns:
        result = re.sub(pattern, replacement, result, flags=re.IGNORECASE)

    # Replace author/name placeholders only if author is provided
    if author and author != "[fullname]":
        name_patterns = [
            r"\[fullname\]",
            r"\[name of copyright owner\]",
            r"\[name\]",
        ]

        for pattern in name_patterns:
            result = re.sub(pattern, author, result, flags=re.IGNORECASE)

    return result


def license_body(license=None, search_name_and_spdx_id=True, refresh=False):
    _license_info = license_info(license, search_name_and_spdx_id, refresh=refresh)
    if _license_info is not None:
        return _license_info["body"]


def license_info(license=None, search_name_and_spdx_id=True, refresh=False):
    licenses = licenses_dict(refresh=refresh)
    if license not in licenses:
        if search_name_and_spdx_id:
            for ll in licenses.values():
                if license in {ll["name"], ll["spdx_id"]}:
                    return ll
        print("That's not a valid license key. Here is a list of valid license keys:")
        print("\t", *licenses, sep="\n\t")
    else:
        return licenses[license]


def licenses_dict(refresh=False):
    licenses = get_licenses(refresh=refresh)
    return {x["key"]: x for x in licenses}


def get_licenses(refresh=False):
    if refresh:
        try:
            licenses = get_licenses_from_github()
            json.dump(licenses, open(licenses_json_path, "w"))
        except Exception:
            raise

    return json.load(open(licenses_json_path))


def get_licenses_from_github():
    """get_licenses_json_from_github
    You need to have a github token placed in the right place for this!
    See pygithub for details.
    ```
    license_jsons = get_licenses_json_from_github()
    ```
    """

    from github import Github

    def gen():
        g = Github()
        license_getter = g.get_licenses()
        i = 0
        while True:
            more = license_getter.get_page(i)
            i += 1
            if len(more) > 0:
                yield more
            else:
                break

    from itertools import chain

    licenses = list(chain.from_iterable(gen()))
    licenses = [ll.raw_data for ll in licenses]
    return licenses



"""
Migration tools for converting old setuptools/CI configurations to modern formats.

This module provides functions to migrate:
- setup.cfg files to pyproject.toml (hatching format)
- Old GitHub CI workflows to modern 2025 format

Key Functions:
    migrate_setuptools_to_hatching: Convert setup.cfg to pyproject.toml
    migrate_github_ci_old_to_new: Convert old CI scripts to new format

Example:
    >>> from wads.migration import migrate_setuptools_to_hatching
    >>> # From a file (use actual file path)
    >>> pyproject = migrate_setuptools_to_hatching('setup.cfg')  # doctest: +SKIP
    >>>
    >>> # From a dict with complete metadata
    >>> cfg = {
    ...     'metadata': {
    ...         'name': 'myproject',
    ...         'version': '0.1.0',
    ...         'description': 'A sample project',
    ...         'url': 'https://github.com/user/myproject',
    ...         'license': 'MIT'
    ...     }
    ... }
    >>> result = migrate_setuptools_to_hatching(cfg)
    >>> 'name = "myproject"' in result
    True
"""

import os
import sys
from typing import Union, Mapping, Callable, Optional
from pathlib import Path
from configparser import ConfigParser
from io import StringIO

# Import toml reading/writing utilities
if sys.version_info >= (3, 11):
    import tomllib
else:
    try:
        import tomli as tomllib
    except ImportError:
        raise ImportError(
            "tomli is required for Python < 3.11. Install with: pip install tomli"
        )

try:
    import tomli_w
except ImportError:
    raise ImportError(
        "tomli_w is required for writing TOML. Install with: pip install tomli_w"
    )

from wads import (
    pyproject_toml_tpl_path,
    github_ci_tpl_publish_path,
    github_ci_publish_2025_path,
)


# --------------------------------------------------------------------------------------
# Setup.cfg to pyproject.toml migration
# --------------------------------------------------------------------------------------


class MigrationError(ValueError):
    """Raised when migration cannot be completed due to missing or unmapped data."""

    pass


def _parse_setup_cfg_string(content: str) -> dict:
    """Parse setup.cfg content string into a dictionary."""
    parser = ConfigParser()
    parser.read_string(content)

    result = {}
    for section in parser.sections():
        result[section] = dict(parser.items(section))
    return result


def _read_setup_cfg_file(filepath: str) -> dict:
    """Read and parse a setup.cfg file."""
    parser = ConfigParser()
    parser.read(filepath)

    result = {}
    for section in parser.sections():
        result[section] = dict(parser.items(section))
    return result


def _normalize_setup_cfg_input(setup_cfg: Union[str, Mapping]) -> dict:
    """
    Normalize setup_cfg input to a dictionary.

    Args:
        setup_cfg: Either a file path, file content string, or dict

    Returns:
        Dictionary representation of setup.cfg contents

    >>> cfg = _normalize_setup_cfg_input({'metadata': {'name': 'test'}})
    >>> cfg['metadata']['name']
    'test'
    """
    if isinstance(setup_cfg, Mapping):
        return dict(setup_cfg)

    # It's a string - either path or content
    if os.path.isfile(setup_cfg):
        return _read_setup_cfg_file(setup_cfg)
    else:
        # Assume it's the content itself
        return _parse_setup_cfg_string(setup_cfg)


def _extract_metadata_value(cfg: dict, *keys, default=None):
    """
    Extract value from setup.cfg dict trying multiple possible keys.

    >>> cfg = {'metadata': {'name': 'myproject'}}
    >>> _extract_metadata_value(cfg, 'name')
    'myproject'
    """
    metadata = cfg.get("metadata", {})
    for key in keys:
        if key in metadata:
            return metadata[key]
    return default


def _extract_options_value(cfg: dict, *keys, default=None):
    """Extract value from [options] section trying multiple possible keys."""
    options = cfg.get("options", {})
    for key in keys:
        if key in options:
            value = options[key]
            # Handle multi-line values (common in install_requires, keywords, etc.)
            if isinstance(value, str) and "\n" in value:
                # Split and clean
                items = [
                    line.strip() for line in value.strip().split("\n") if line.strip()
                ]
                return items if items else default
            return value
    return default


def _parse_list_field(value: Optional[str]) -> list:
    """Parse a multi-line or comma-separated string field into a list."""
    if not value:
        return []

    if isinstance(value, list):
        return value

    # Handle multi-line format (common in setup.cfg)
    if "\n" in value:
        return [line.strip() for line in value.strip().split("\n") if line.strip()]

    # Handle comma-separated format
    if "," in value:
        return [item.strip() for item in value.split(",") if item.strip()]

    # Single value
    return [value.strip()] if value.strip() else []


# Transformation rules: each rule takes the setup.cfg dict and returns the value for that field
def _rule_project_name(cfg: dict) -> str:
    """Extract project name."""
    return _extract_metadata_value(cfg, "name")


def _rule_project_version(cfg: dict) -> str:
    """Extract project version."""
    return _extract_metadata_value(cfg, "version")


def _rule_project_description(cfg: dict) -> str:
    """Extract project description."""
    return _extract_metadata_value(cfg, "description", default="")


def _rule_project_url(cfg: dict) -> str:
    """Extract project URL."""
    return _extract_metadata_value(cfg, "url", "home_page")


def _rule_project_license(cfg: dict) -> str:
    """Extract license identifier."""
    license_value = _extract_metadata_value(cfg, "license")
    if license_value:
        # Normalize common license names
        license_map = {
            "mit": "MIT",
            "apache-2.0": "Apache-2.0",
            "apache software license": "Apache-2.0",
            "bsd": "BSD",
            "gpl": "GPL",
        }
        return license_map.get(license_value.lower(), license_value)
    return license_value


def _rule_project_keywords(cfg: dict) -> list:
    """Extract keywords as list."""
    keywords = _extract_metadata_value(cfg, "keywords", default="")
    return _parse_list_field(keywords)


def _rule_project_authors(cfg: dict) -> list:
    """Extract authors list."""
    author = _extract_metadata_value(cfg, "author")
    author_email = _extract_metadata_value(cfg, "author_email", "author-email")

    if author or author_email:
        author_dict = {}
        if author:
            author_dict["name"] = author
        if author_email:
            author_dict["email"] = author_email
        return [author_dict]
    return []


def _rule_project_dependencies(cfg: dict) -> list:
    """Extract install_requires as dependencies."""
    deps = _extract_options_value(cfg, "install_requires", default="")
    parsed = _parse_list_field(deps)
    # Filter out placeholder/invalid dependencies
    return [d for d in parsed if d and not d.startswith("this_does_not_exist")]


def _rule_project_optional_dependencies(cfg: dict) -> dict:
    """Extract extras_require as optional-dependencies."""
    extras = cfg.get("options.extras_require", {})
    if not extras:
        return {}

    result = {}
    for key, value in extras.items():
        result[key] = _parse_list_field(value)
    return result


def _rule_project_entry_points(cfg: dict) -> dict:
    """Extract console_scripts from entry_points."""
    entry_points = cfg.get("options.entry_points", {})
    if not entry_points:
        return {}

    console_scripts = entry_points.get("console_scripts", "")
    if not console_scripts:
        return {}

    scripts = _parse_list_field(console_scripts)
    if scripts:
        return {"console_scripts": scripts}
    return {}


# Global rules dictionary mapping field paths to extraction functions
setup_cfg_to_pyproject_toml_rules = {
    "project.name": _rule_project_name,
    "project.version": _rule_project_version,
    "project.description": _rule_project_description,
    "project.url": _rule_project_url,
    "project.license": _rule_project_license,
    "project.keywords": _rule_project_keywords,
    "project.authors": _rule_project_authors,
    "project.dependencies": _rule_project_dependencies,
    "project.optional-dependencies": _rule_project_optional_dependencies,
    "project.scripts": _rule_project_entry_points,
}


def _load_pyproject_template() -> str:
    """Load the pyproject.toml template."""
    with open(pyproject_toml_tpl_path, "r") as f:
        return f.read()


def _apply_rules_to_extract_values(cfg: dict, rules: dict) -> dict:
    """
    Apply extraction rules to setup.cfg dict to get pyproject.toml values.

    Returns a nested dictionary structure matching pyproject.toml layout.
    """
    result = {}

    for field_path, rule_func in rules.items():
        value = rule_func(cfg)
        if value is not None and value != [] and value != {}:
            # Build nested dict structure from dot-separated path
            parts = field_path.split(".")
            current = result
            for part in parts[:-1]:
                if part not in current:
                    current[part] = {}
                current = current[part]
            current[parts[-1]] = value

    return result


def migrate_setuptools_to_hatching(
    setup_cfg: Union[str, Mapping],
    defaults: Optional[dict] = None,
    *,
    rules: Optional[dict] = None,
) -> str:
    """
    Migrate setup.cfg to pyproject.toml format using hatching.

    Args:
        setup_cfg: Either a file path, file content string, or dict of setup.cfg
        defaults: Default values to use for missing required fields
        rules: Custom transformation rules (defaults to setup_cfg_to_pyproject_toml_rules)

    Returns:
        String content of the generated pyproject.toml

    Raises:
        MigrationError: If required fields are missing or unmapped data exists

    Example:
        >>> # From file path
        >>> pyproject = migrate_setuptools_to_hatching('setup.cfg')  # doctest: +SKIP
        >>>
        >>> # From dict with all required fields
        >>> cfg = {
        ...     'metadata': {
        ...         'name': 'myproj',
        ...         'version': '0.1.0',
        ...         'description': 'My project',
        ...         'url': 'https://github.com/user/myproj',
        ...         'license': 'MIT'
        ...     }
        ... }
        >>> result = migrate_setuptools_to_hatching(cfg)
        >>> 'name = "myproj"' in result
        True
        >>>
        >>> # With defaults for missing fields
        >>> minimal = {'metadata': {'name': 'test', 'version': '1.0'}}
        >>> result = migrate_setuptools_to_hatching(
        ...     minimal,
        ...     defaults={
        ...         'description': 'Test project',
        ...         'url': 'https://test.com',
        ...         'license': 'MIT'
        ...     }
        ... )
        >>> 'name = "test"' in result
        True
    """
    if defaults is None:
        defaults = {}

    if rules is None:
        rules = setup_cfg_to_pyproject_toml_rules

    # Normalize input to dict
    cfg = _normalize_setup_cfg_input(setup_cfg)

    # Load template
    template = _load_pyproject_template()

    # Apply rules to extract values
    extracted = _apply_rules_to_extract_values(cfg, rules)

    # Required placeholders in template
    required_fields = {
        "name": extracted.get("project", {}).get("name") or defaults.get("name"),
        "version": extracted.get("project", {}).get("version")
        or defaults.get("version"),
        "description": extracted.get("project", {}).get("description")
        or defaults.get("description", ""),
        "url": extracted.get("project", {}).get("url") or defaults.get("url", ""),
        "license": extracted.get("project", {}).get("license")
        or defaults.get("license", "MIT"),
    }

    # Check for missing required fields
    missing = [k for k, v in required_fields.items() if not v]
    if missing:
        raise MigrationError(
            f"Missing required fields: {missing}. "
            f"Provide them in setup.cfg or via the defaults parameter."
        )

    # Parse the template as TOML (it has placeholders but is valid TOML structure)
    pyproject_dict = tomllib.loads(template)

    # Replace placeholder values with actual values
    if "project" not in pyproject_dict:
        pyproject_dict["project"] = {}

    pyproject_dict["project"]["name"] = required_fields["name"]
    pyproject_dict["project"]["version"] = required_fields["version"]
    pyproject_dict["project"]["description"] = required_fields["description"]
    pyproject_dict["project"]["license"] = {"text": required_fields["license"]}

    if "urls" not in pyproject_dict["project"]:
        pyproject_dict["project"]["urls"] = {}
    pyproject_dict["project"]["urls"]["Homepage"] = required_fields["url"]

    # Merge extracted values
    project_section = extracted.get("project", {})

    # Add keywords if present
    if "keywords" in project_section and project_section["keywords"]:
        pyproject_dict["project"]["keywords"] = project_section["keywords"]

    # Add authors if present
    if "authors" in project_section and project_section["authors"]:
        pyproject_dict["project"]["authors"] = project_section["authors"]

    # Add dependencies if present
    if "dependencies" in project_section and project_section["dependencies"]:
        pyproject_dict["project"]["dependencies"] = project_section["dependencies"]

    # Add optional dependencies if present
    if (
        "optional-dependencies" in project_section
        and project_section["optional-dependencies"]
    ):
        pyproject_dict["project"]["optional-dependencies"] = project_section[
            "optional-dependencies"
        ]

    # Add scripts/entry points if present
    if "scripts" in project_section and project_section["scripts"]:
        if "console_scripts" in project_section["scripts"]:
            pyproject_dict["project"]["scripts"] = {}
            for script in project_section["scripts"]["console_scripts"]:
                if "=" in script:
                    name, target = script.split("=", 1)
                    pyproject_dict["project"]["scripts"][name.strip()] = target.strip()

    # Convert back to TOML string
    from io import BytesIO

    output = BytesIO()
    # Write with tomli_w for consistent formatting
    tomli_w.dump(pyproject_dict, output)

    return output.getvalue().decode("utf-8")


# --------------------------------------------------------------------------------------
# GitHub CI migration
# --------------------------------------------------------------------------------------


def _load_ci_template(filepath: str) -> str:
    """Load CI template file."""
    with open(filepath, "r") as f:
        return f.read()


def _extract_project_name_from_ci(old_ci_content: str) -> Optional[str]:
    """Extract PROJECT_NAME from old CI script."""
    for line in old_ci_content.split("\n"):
        if "PROJECT_NAME:" in line and "#PROJECT_NAME#" in line:
            return None  # Placeholder not filled
        elif "PROJECT_NAME:" in line:
            parts = line.split(":", 1)
            if len(parts) == 2:
                name = parts[1].strip()
                # Remove comments
                if "#" in name:
                    name = name.split("#")[0].strip()
                return name
    return None


def migrate_github_ci_old_to_new(
    old_ci: Union[str, Path],
    defaults: Optional[dict] = None,
) -> str:
    """
    Migrate old GitHub CI script to new 2025 format.

    Args:
        old_ci: Path to old CI file or its content as string
        defaults: Default values for missing fields (e.g., {'project_name': 'myproject'})

    Returns:
        String content of the new CI script

    Raises:
        MigrationError: If unmapped elements exist or required fields are missing

    Example:
        >>> # From file
        >>> new_ci = migrate_github_ci_old_to_new('.github/workflows/ci.yml')  # doctest: +SKIP
        >>>
        >>> # From content string with project name
        >>> old_ci = '''
        ... name: CI
        ... env:
        ...   PROJECT_NAME: myproject
        ... '''
        >>> result = migrate_github_ci_old_to_new(old_ci)
        >>> 'PROJECT_NAME: myproject' in result
        True
        >>>
        >>> # With defaults when project name is missing
        >>> old_ci_minimal = '''
        ... name: CI
        ... env:
        ...   PROJECT_NAME: #PROJECT_NAME#
        ... '''
        >>> result = migrate_github_ci_old_to_new(
        ...     old_ci_minimal,
        ...     defaults={'project_name': 'test'}
        ... )
        >>> 'PROJECT_NAME: test' in result
        True
    """
    if defaults is None:
        defaults = {}

    # Read old CI content
    if isinstance(old_ci, Path):
        old_ci = str(old_ci)

    if os.path.isfile(old_ci):
        with open(old_ci, "r") as f:
            old_content = f.read()
    else:
        old_content = old_ci

    # Load new template
    new_template = _load_ci_template(github_ci_publish_2025_path)

    # Extract project name
    project_name = _extract_project_name_from_ci(old_content)
    if not project_name:
        project_name = defaults.get("project_name")

    if not project_name:
        raise MigrationError(
            "Could not extract PROJECT_NAME from old CI script. "
            "Provide it via defaults={'project_name': 'yourproject'}"
        )

    # Check for elements in old CI that might not be in new template
    # This is a basic check - you can expand this based on specific needs
    old_has_setuptools = "setuptools" in old_content.lower()
    old_has_pylint = "pylint" in old_content.lower()

    warnings = []
    if old_has_setuptools:
        warnings.append("Old CI uses setuptools - ensure pyproject.toml is ready")
    if old_has_pylint:
        warnings.append("Old CI uses pylint - new CI uses ruff for linting")

    # Replace project name placeholder
    result = new_template.replace("#PROJECT_NAME#", project_name)

    # Add warnings as comments if any
    if warnings:
        warning_text = "\n".join(f"# MIGRATION NOTE: {w}" for w in warnings)
        # Insert after the first line (name: ...)
        lines = result.split("\n")
        lines.insert(1, warning_text)
        result = "\n".join(lines)

    return result



"""Tests for migration module."""

import os
import tempfile
from pathlib import Path

import pytest

from wads.migration import (
    migrate_setuptools_to_hatching,
    migrate_github_ci_old_to_new,
    MigrationError,
    _normalize_setup_cfg_input,
    _parse_list_field,
)


# Sample setup.cfg content for testing
SAMPLE_SETUP_CFG = """
[metadata]
name = testproject
version = 0.1.0
url = https://github.com/test/testproject
description = A test project
license = MIT
keywords =
    testing
    migration
author = Test Author
author_email = test@example.com

[options]
packages = find:
include_package_data = True
zip_safe = False
install_requires =
    requests
    click>=7.0

[options.extras_require]
dev =
    pytest
    black

[options.entry_points]
console_scripts =
    testcli = testproject.cli:main
"""


def test_normalize_setup_cfg_dict():
    """Test normalizing dict input."""
    cfg_dict = {"metadata": {"name": "test"}}
    result = _normalize_setup_cfg_input(cfg_dict)
    assert result == cfg_dict


def test_normalize_setup_cfg_string():
    """Test normalizing string content input."""
    result = _normalize_setup_cfg_input(SAMPLE_SETUP_CFG)
    assert "metadata" in result
    assert result["metadata"]["name"] == "testproject"


def test_parse_list_field_multiline():
    """Test parsing multi-line list fields."""
    field = "\n    testing\n    migration\n"
    result = _parse_list_field(field)
    assert result == ["testing", "migration"]


def test_parse_list_field_comma_separated():
    """Test parsing comma-separated list fields."""
    field = "testing, migration, tools"
    result = _parse_list_field(field)
    assert result == ["testing", "migration", "tools"]


def test_parse_list_field_empty():
    """Test parsing empty field."""
    assert _parse_list_field("") == []
    assert _parse_list_field(None) == []


def test_migrate_setuptools_basic():
    """Test basic migration from setup.cfg to pyproject.toml."""
    result = migrate_setuptools_to_hatching(SAMPLE_SETUP_CFG)

    # Check it's valid TOML-ish output
    assert 'name = "testproject"' in result
    assert 'version = "0.1.0"' in result
    assert 'description = "A test project"' in result
    # Keywords should be present (format may vary)
    assert '"testing"' in result
    assert '"migration"' in result


def test_migrate_setuptools_with_defaults():
    """Test migration with defaults for missing fields."""
    minimal_cfg = """
[metadata]
name = minimal
version = 1.0.0
"""

    defaults = {
        "description": "A minimal project",
        "url": "https://example.com",
        "license": "MIT",
    }

    result = migrate_setuptools_to_hatching(minimal_cfg, defaults=defaults)
    assert 'name = "minimal"' in result
    assert "A minimal project" in result


def test_migrate_setuptools_missing_required_fields():
    """Test that missing required fields raise MigrationError."""
    minimal_cfg = """
[metadata]
name = minimal
"""

    with pytest.raises(MigrationError) as exc_info:
        migrate_setuptools_to_hatching(minimal_cfg)

    assert "Missing required fields" in str(exc_info.value)
    assert "version" in str(exc_info.value)


def test_migrate_setuptools_from_file():
    """Test migration from actual file."""
    with tempfile.NamedTemporaryFile(mode="w", suffix=".cfg", delete=False) as f:
        f.write(SAMPLE_SETUP_CFG)
        temp_path = f.name

    try:
        result = migrate_setuptools_to_hatching(temp_path)
        assert 'name = "testproject"' in result
    finally:
        os.unlink(temp_path)


def test_migrate_setuptools_dependencies():
    """Test that dependencies are properly migrated."""
    result = migrate_setuptools_to_hatching(SAMPLE_SETUP_CFG)

    # Check dependencies section (format may vary)
    assert "dependencies" in result
    assert '"requests"' in result or "requests" in result
    assert "click" in result


def test_migrate_ci_missing_project_name():
    """Test CI migration raises error when project name is missing."""
    old_ci = """
name: Old CI
env:
  PROJECT_NAME: #PROJECT_NAME#
"""

    with pytest.raises(MigrationError) as exc_info:
        migrate_github_ci_old_to_new(old_ci)

    assert "PROJECT_NAME" in str(exc_info.value)


def test_migrate_ci_with_defaults():
    """Test CI migration with defaults."""
    old_ci = """
name: Old CI
env:
  PROJECT_NAME: #PROJECT_NAME#
"""

    result = migrate_github_ci_old_to_new(
        old_ci, defaults={"project_name": "myproject"}
    )

    assert "PROJECT_NAME: myproject" in result


def test_migrate_ci_extracts_project_name():
    """Test CI migration extracts existing project name."""
    old_ci = """
name: Old CI
env:
  PROJECT_NAME: existing_project
"""

    result = migrate_github_ci_old_to_new(old_ci)
    assert "PROJECT_NAME: existing_project" in result


def test_custom_rules():
    """Test using custom transformation rules."""
    custom_rules = {
        "project.name": lambda cfg: cfg.get("metadata", {}).get("name", "").upper(),
        "project.version": lambda cfg: cfg.get("metadata", {}).get("version"),
        "project.description": lambda cfg: "Custom description",
        "project.url": lambda cfg: "https://custom.url",
        "project.license": lambda cfg: "MIT",
    }

    cfg_dict = {
        "metadata": {
            "name": "test",
            "version": "1.0.0",
        }
    }

    result = migrate_setuptools_to_hatching(cfg_dict, rules=custom_rules)
    assert 'name = "TEST"' in result
    assert "Custom description" in result
